# -*- coding: utf-8 -*-
from unittest import TestCase
from nose.plugins.attrib import attr
import os
from copydog.api.trello import Trello, Board, Card, List, Member


@attr('http', 'trello')
class TestTrello(TestCase):
    """ Make sure to provide ``TRELLO_API_KEY`` and ``TRELLO_TOKEN``
        environment variables to execute this test.
    """

    def setUp(self):
        self.trello = Trello(api_key=os.environ.get('COPYDOG_TRELLO_API_KEY'),
                             token=os.environ.get('COPYDOG_TRELLO_TOKEN'),
        )
        self.board_id = '4d5ea62fd76aa1136000000c'

    def test_trello_connect(self):
        boards = self.trello.boards()
        self.assertTrue(next(boards, Board()), Board)

    def test_trello_cards(self):
        cards = self.trello.cards(board_id=self.board_id)
        self.assertTrue(next(cards), Card)

    def test_trello_lists(self):
        lists = self.trello.lists(board_id=self.board_id)
        self.assertTrue(next(lists), List)

    def test_trello_members(self):
        members = self.trello.lists(board_id=self.board_id)
        self.assertTrue(next(members), Member)


# -*- coding: utf-8 -*-
from unittest import TestCase
from mock import patch
from nose.plugins.attrib import attr
from copydog.utils.pandoc import convert


@attr('pandoc')
class PandocTest(TestCase):

    def test_pandoc(self):
        markdown_text = u"""\
Hello
=====
We're loving Markdown"""

        expected_result = u"""\
h1. Hello

We're loving Markdown

"""

        self.assertEqual(convert(markdown_text, 'markdown', 'textile'), expected_result)

    def test_unicode(self):
        unicode_text = u'Привет'
        self.assertEqual(convert(unicode_text, 'textile', 'markdown').strip(), unicode_text)

    @patch('copydog.utils.pandoc.PANDOC_PATH', '/tmp/')
    def test_not_installed(self):
        unicode_text = u'Hello _worlds_'
        self.assertEqual(convert(unicode_text, 'textile', 'markdown').strip(), unicode_text)


# -*- coding: utf-8 -*-
"""
Copydog syncs Redmine and Trello.

Usage:
  runner.py --config=<yaml> [options]
  runner.py (start|stop|restart) --config=<yaml> [options]
  runner.py (debug_storage|flush_storage) [options]

Options:
  --config=<yaml>  Config file.
  -f --fullsync    Full sync (all issues, opposed to date range delta sync). Needed only once.
  -v --verbose     Make verbose output.
  -q --quiet       Make no output.
  -h --help        Show this screen.
  --version        Show version.
"""

import logging
import logging.config
from daemon.runner import DaemonRunner
from docopt import docopt

import copydog
from ..storage.factory import StorageFactory
from ..utils.config import Config
from ..utils import storage_browser
from ..watcher import Watch


def setup_logging(arguments):
    logging.config.fileConfig('logging.cfg', disable_existing_loggers=True)
    if arguments['--verbose']:
        level = logging.DEBUG
    elif arguments['--quiet']:
        level = logging.CRITICAL
    else:
        level = logging.INFO
    logging.getLogger('copydog').setLevel(level)


def setup_config(config_path):
    config = Config()
    try:
        config = Config(file=config_path)
    except Exception as e:
        exit(str(e))
    return config


class DeamonApp(object):
    stdin_path='/tmp/copydog.in.log'
    stdout_path='/tmp/copydog.out.log'
    stderr_path='/tmp/copydog.err.log'
    pidfile_path='/tmp/copydog.pid'
    pidfile_timeout=100
    run=lambda: None


def execute(config, full_sync=False, daemonize=False):
    if not any(map(lambda item: item[1].get('write'), config.clients)):
        exit('Allow at least one client to write')

    watch = Watch(config, full_sync=full_sync)

    if daemonize:
        app = DeamonApp()
        app.run = watch.run
        DaemonRunner(app).do_action()
    else:
        watch.run()


def flush_storage(storage):
    """ Erase all copydog keys and values from storage"""
    msg = 'This is irreversible operation, please confirm you want to empty copydog storage.'
    shall = True if raw_input("%s (y/N) " % msg).lower() == 'y' else False
    if shall:
        storage.flush()
    else:
        print 'No action taken'


def main():
    arguments = docopt(__doc__, version='Copydog %s' % copydog.__version__)
    setup_logging(arguments)
    config = setup_config(arguments['--config'])
    storage = StorageFactory.get(config.get('storage'))

    if arguments['debug_storage']:
        storage_browser.browse(storage)
        exit()

    if arguments['flush_storage']:
        flush_storage(storage)
        exit()

    daemonize = bool(arguments.get('start') or arguments.get('stop') or arguments.get('restart'))
    execute(config, full_sync=arguments['--fullsync'], daemonize=daemonize)

if __name__ == '__main__':
    try:
        main()
    except KeyboardInterrupt:
        pass



# -*- coding: utf-8 -*-
import mock
import os
import unittest
if not hasattr(unittest.TestCase, 'assertRaisesRegexp'):
    import unittest2 as unittest
from copydog.utils.config import Config, ImproperlyConfigured, MissigAttr


class ConfigInitTest(unittest.TestCase):
    example_yml = 'examples/copydog.yml'

    def test_empty_config_init(self):
        config = Config()
        self.assertEqual(config.get('option'), None)

    def test_config_init_dict(self):
        config = Config({'option': 'value'})
        self.assertEqual(config.get('option'), 'value')
        self.assertEqual(config.get('nonexistent'), None)
        self.assertEqual(config.get('nonexistent', False), False)

    def test_config_init_yml(self):
        config = Config(file=self.example_yml)
        self.assertTrue(type(config.get('clients')) is dict)
        self.assertSequenceEqual(config.get('clients').keys(), ('redmine', 'trello'))

    def test_bad_init(self):
        self.assertRaisesRegexp(ImproperlyConfigured, 'Provide either data or filepath',
                                Config, data={'option': 'value'}, file=self.example_yml)

def kwargs_test(**kwargs):
    return True


class ConfigAttrTest(unittest.TestCase):
    example_yml = 'examples/copydog.yml'

    def setUp(self):
        self.config = Config(file=self.example_yml)

    def test_attrs_scalar(self):
        self.assertEqual(self.config.clients.redmine.write, 1)

    def test_attr_subconfig(self):
        self.assertTrue(type(self.config.clients.trello), Config)
        self.assertTrue(kwargs_test(**self.config))

    def test_attr_nonexistent(self):
        with self.assertRaisesRegexp(MissigAttr, 'Missing config item clients.redmine.long_attr'):
            me = self.config.clients.redmine.long_attr

    def test_set_simple_attr(self):
        self.config.additional = 5
        self.assertEqual(self.config.additional, 5)
        self.assertEqual(self.config.get('additional'), None)

    def test_name(self):
        self.assertEqual(str(self.config), '')
        self.assertEqual(str(self.config.clients.redmine), 'clients.redmine')

    def test_iterable(self):
        for service, config in self.config.clients:
            self.assertTrue(type(service) is str)
            self.assertTrue(type(config) is Config)

    @mock.patch.dict('os.environ', {'COPYDOG_REDMINE_API_KEY': 'XXX'})
    def test_env_fallback(self):
        self.assertEqual(self.config.clients.redmine.api_key, 'XXX')



# -*- coding: utf-8 -*-
from exceptions import ImportError, Exception


class StorageFactory(object):
    default_backend = 'redis'

    @classmethod
    def get_class_name(cls, storage_name):
        if '.' not in storage_name:
            module_name = 'copydog.storage.backends.{0}_backend'.format(storage_name)
            class_name = 'Storage'
        else:
            module_name = storage_name[:storage_name.rfind(".")]
            class_name = storage_name[storage_name.rfind(".") + 1:]
        module = __import__(module_name, globals(), locals(), [class_name, ])
        storage_klass = getattr(module, class_name)
        return storage_klass

    @classmethod
    def get(cls, storage_config):
        try:
            if storage_config:
                storage_name = storage_config.keys()[0]
                options = storage_config[storage_name]
            else:
                storage_name = cls.default_backend
                options = None
            storage_klass = cls.get_class_name(storage_name)
            return storage_klass(storage_options=options)
        except ImportError as e:
            raise Exception('Copydog storage in not configured properly: %s' % str(e))


# -*- coding: utf-8 -*-
import logging
import subprocess
log = logging.getLogger('copydog')


PANDOC_PATH = '/usr/bin/pandoc'


def convert(text, from_format, to_format):
    """ Convert markups using pandoc
        Inspired by https://github.com/kennethreitz/pyandoc

        In case of error returns input text as output.
    """
    try:
        p = subprocess.Popen(
            [PANDOC_PATH, '--from=%s' % from_format, '--to=%s' % to_format],
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE
        )
        return p.communicate(text.encode('utf-8'))[0].decode('utf-8')
    except OSError as e:
        log.warn('Pandoc is misconfigured, %s', str(e))
        return text
    except Exception:
        return text




# -*- coding: utf-8 -*-
from logging import getLogger
import time
import sched
import itertools
from copydog.utils import pandoc
from copydog.adapters import RedmineAdapter, TrelloAdapter
from copydog.convertor import Mapper
from copydog.storage.factory import StorageFactory
from copydog.utils.task import periodic


log = getLogger('copydog')


class Watch(object):
    available_services = {
        'redmine': RedmineAdapter,
        'trello': TrelloAdapter
    }

    def __init__(self, config=None, full_sync=None):
        log.info('Copydog is on duty...')
        if config:
            self.setup(config, full_sync)

    def setup(self, config, full_sync=False):
        self.storage = StorageFactory.get(config.get('storage'))
        self.services = self.setup_services(config, self.storage)
        self.setup_copydog(config.get('copydog'))
        self.setup_last_time_read(full_sync)
        self.mapper = Mapper(storage=self.storage, services=self.services, config=config)
        self.mapper.save_list_status_mapping()
        self.mapper.save_user_member_mapping()

    def setup_copydog(self, options=False):
        if not options:
            return
        pandoc.PANDOC_PATH = options.get('pandoc', '/usr/bin/pandoc')
        self.beat_interval = options.get('beat', 60)

    def setup_services(self, config, storage):
        clients_config = config.clients
        services = {}
        for service_name, options in clients_config:
            service_class = self.available_services[service_name]
            services[service_name] = service_class(options, storage)
        return services

    def setup_last_time_read(self, full_sync=False):
        """ Ignoring last time read, when using full_sync,
            If launching for first time, make sure we're monitoring only recent changes.

            TODO: remove hardcoded services
        """
        if full_sync:
            self.storage.reset_last_time_read()
        else:
            if not self.storage.get_last_time_read('redmine'):
                self.storage.mark_read('redmine')
                self.storage.mark_read('trello')

    def run(self):
        """ Run copydog in a loop, starting sync every N seconds (60 is default).
        """
        self.scheduler = sched.scheduler(time.time, time.sleep)
        periodic(self.scheduler, self.beat_interval, self.sync)
        self.scheduler.run()

    def sync(self):
        groups = itertools.groupby(itertools.permutations(self.services, 2), lambda pair:pair[0])
        for service_from_name, services in groups:
            service_from = self.services[service_from_name]
            issues = service_from.read()
            num_issues_read = 0
            services = list(services)
            for issue in issues:
                if self.is_already_synced(issue):
                    continue
                num_issues_read += 1
                service_from.mark_read(issue)
                for service_from_name, service_to_name in services:
                    service_to = self.services[service_to_name]
                    if service_to.writable:
                        service_to.write(issue)
            log.info('Read %s new issues from %s', num_issues_read, service_from)

    def is_already_synced(self, issue):
        """ Check issue needs sync

            Read from API last_updated > read from storage last_updated
            thus preventing clones
        """
        last_time_synced = self.storage.get_last_time_updated(issue.client.service_name, issue)
        log.debug('Comparing %s <= %s (%s)' % (issue.last_updated, last_time_synced,
            (last_time_synced and issue.last_updated <= last_time_synced)))
        return last_time_synced and issue.last_updated <= last_time_synced




# -*- coding: utf-8 -*-
import os
import yaml

class ImproperlyConfigured(Exception):
    """ Something wrong with config file
    """

class MissigAttr(AttributeError):
    """ Config doesn't have specific value
    """

class Config(object):
    """ Object-like config with dict inside.
    """

    def __init__(self, data=None, file=None, name=''):
        if data and file:
            raise ImproperlyConfigured('Provide either data or filepath')
        if file:
            data = yaml.load(open(file))
        self.__data = data or {}
        self.__name = name

    def get(self, key, value=None):
        return self.__data.get(key, value)

    def __getattr__(self, name):
        try:
            attr = self.__data[name]
        except KeyError:
            return self.__getfallback(name)

        if type(attr) == dict:
            # creating subconfig
            return Config(attr, name=self.__getname(name))
        else:
            return attr

    def __getname(self, name):
        return '.'.join((self.__name, name)) if self.__name else name

    def __getfallback(self, name):
        parts = self.__name.split('.')
        parts.append(name)
        parts = ['copydog'] + parts[-2:]
        env_var = '_'.join(parts).upper()
        if env_var in os.environ:
            return os.environ.get(env_var)
        raise MissigAttr('Missing config item %s' % self.__getname(name))

    def __str__(self):
        return self.__name

    def __iter__(self):
        for key in self.__data.iterkeys():
            yield key, self.__getattr__(key)

    def keys(self):
        return self.__data.keys()

    def __getitem__(self, key):
        return self.__getattr__(key)



# -*- coding: utf-8 -*-
from unittest import TestCase
import datetime
from mock import Mock, patch
from copydog.adapters import RedmineAdapter, TrelloAdapter
from copydog.api.redmine import Issue
from copydog.convertor import Mapper
from copydog.api.trello import Card
from copydog.utils.config import Config
config = Config(file='tests/test.yml')
redmine_options = config.clients.redmine
trello_options = config.clients.trello


class IssueToCardTest(TestCase):

    def setUp(self):
        self.storage_mock = Mock()
        self.adapter = TrelloAdapter(trello_options, self.storage_mock)
        self.issue = Issue(**{
            'id': 6571,
            'subject': 'Save humanity',
            'description': 'Humanity in danger',
            'due_date': datetime.date(2012, 12, 20),
            'status': {'name': "Assigned", 'id': 2},
            'assigned_to': {'id': '42'}
        })

    @property
    def trello_card(self):
        return self.adapter.convert_to_local_issue(self.issue)

    def test_new_card_id(self):
        self.storage_mock.get_opposite_item_id.return_value = '777'
        self.assertEqual(self.trello_card.id, '777')

    def test_card_idMembers(self):
        self.storage_mock.get_user_or_member_id.return_value = 'abcdef'
        self.assertEqual(self.trello_card.idMembers, ['abcdef'])

    def test_card_name(self):
        self.assertEqual(self.trello_card.name, 'Save humanity')

    def test_card_desc(self):
        self.assertEqual(self.trello_card.desc.strip(), 'Humanity in danger')

    def test_id_list(self):
        self.storage_mock.get_list_or_status_id.return_value = '123'
        self.assertEqual(self.trello_card.idList, '123')

    @patch.object(trello_options, 'board_id', '1fe889e4c23b476f4a189ca5')
    def test_idBoard(self):
        self.assertEqual(self.trello_card.idBoard, '1fe889e4c23b476f4a189ca5')

    def test_due(self):
        self.assertEqual(self.trello_card.due, datetime.date(2012, 12, 20))

    def test_client(self):
        self.assertEqual(self.trello_card.client.service_name, 'trello')






class CardToIssueTest(TestCase):
    def setUp(self):
        self.storage_mock = Mock()
        options = Config({'host': 'xxx', 'api_key': 'yyy'})
        self.adapter = RedmineAdapter(redmine_options, self.storage_mock)
        self.issue = Card(**{
            'id': 'a84d79f572fbe7512b999c6b3cd7667cbe3138ff',
            'name': 'Destroy humanity',
            'desc': 'Humanity in our enemy',
            'due': datetime.date(1990, 1, 1),
            'idList': '4fe889e4c23b476f4a189ca6',
            'idMembers': [None],
        })

    @property
    def redmine_issue(self):
        return self.adapter.convert_to_local_issue(self.issue)

    def test_new_issue_id(self):
        self.storage_mock.get_opposite_item_id.return_value = '555'
        self.assertEqual(self.redmine_issue.id, '555')

    def test_issue_assigned_to(self):
        self.storage_mock.get_user_or_member_id.return_value = 42
        self.assertEqual(self.redmine_issue.assigned_to_id, 42)

    def test_issue_subject(self):
        self.assertEqual(self.redmine_issue.subject, 'Destroy humanity')

    def test_issue_description(self):
        self.assertEqual(self.redmine_issue.description.strip(), 'Humanity in our enemy')

    def test_status(self):
        self.storage_mock.get_list_or_status_id.return_value = '123'
        self.assertEqual(self.redmine_issue.status_id, '123')

    @patch.object(redmine_options, 'project_id', 7)
    def test_project_id(self):
        self.assertEqual(self.redmine_issue.project_id, 7)

    def test_due_date(self):
        self.assertEqual(self.redmine_issue.due_date, datetime.date(1990, 1, 1))

    def test_client(self):
        self.assertEqual(self.redmine_issue.client.service_name, 'redmine')




# -*- coding: utf-8 -*-
import datetime
import pytz
import redis
from dateutil.parser import parse
from logging import getLogger
from ..base import BaseStorage
log = getLogger('copydog')


class Storage(BaseStorage):
    """ Copydog default sotrage implementation using Redis
    """

    def __init__(self, storage_options=None):
        if not storage_options:
            storage_options  = {}
        self.redis = redis.StrictRedis(**storage_options)

    def get_opposite_item_id(self, service_name, id):
        return self.redis.hget('{service_name}:items:{id}'.format(service_name=service_name, id=id), 'opposite_id')

    def get_list_or_status_id(self, service_name, id):
        return self.redis.hget('{service_name}:list_status_mapping'.format(service_name=service_name), id)

    def set_list_or_status_id(self, redmine_id, trello_id):
        pipe = self.redis.pipeline()
        pipe.hset('redmine:list_status_mapping', redmine_id, trello_id)
        pipe.hset('trello:list_status_mapping', trello_id, redmine_id)
        pipe.execute()

    def get_user_or_member_id(self, service_name, id):
        return self.redis.hget('{service_name}:user_member_mapping'.format(service_name=service_name), id)

    def set_user_or_member_id(self, redmine_id, trello_id):
        """ TODO: extract method, set_list_or_status_id
        """
        pipe = self.redis.pipeline()
        pipe.hset('redmine:user_member_mapping', redmine_id, trello_id)
        pipe.hset('trello:user_member_mapping', trello_id, redmine_id)
        pipe.execute()

    def get_last_time_read(self, service_name):
        value = self.redis.get('{service_name}:last_read_time'.format(service_name=service_name))
        if value:
            return parse(value)
        return None

    def get_last_time_updated(self, service_name, item):
        value = self.redis.hget('{service_name}:items:{id}'.format(service_name=service_name, id=item.id), 'updated')
        if value:
            return parse(value)
        return None

    def reset_last_time_read(self):
        """ TODO: remove hardcoded service names
        """
        log.debug('Resetting last read time for full sync')
        self.redis.delete('{service_name}:last_read_time'.format(service_name='redmine'))
        self.redis.delete('{service_name}:last_read_time'.format(service_name='trello'))

    def mark_read(self, service_name, issue=None):
        """ Marking read whole service if not issue provided.
            In other case mark issue read from API.
        """
        if issue:
            self.redis.hset('{service_name}:items:{id}'.format(service_name=service_name, id=issue.id),
                      'updated', issue.last_updated)
        else:
            self.redis.set('{service_name}:last_read_time'.format(service_name=service_name),
                datetime.datetime.utcnow().replace(tzinfo = pytz.utc))

    def mark_written(self, service_name, item, foreign_id):
        other_service = 'redmine' if service_name == 'trello' else 'trello'
        pipe = self.redis.pipeline()
        pipe.hmset('{service_name}:items:{id}'.format(service_name=service_name, id=item.id),
                  {'opposite_id': foreign_id, 'updated': item.last_updated})
        pipe.hset('{other_service}:items:{id}'.format(other_service=other_service, id=foreign_id),
                  'opposite_id', item.id)
        pipe.execute()

    def flush(self):
        redmine = self.redis.keys(pattern='redmine:*')
        trello = self.redis.keys(pattern='trello:*')
        keys_to_delete = redmine + trello
        if keys_to_delete:
            self.redis.delete(*keys_to_delete)
            log.debug('Deleted keys: %s', keys_to_delete)
            log.info('Deleted %d keys', len(keys_to_delete))
        else:
            log.info('Storage is empty')




# -*- coding: utf-8 -*-
from logging import getLogger

log = getLogger('copydog')


class Mapper(object):
    """
        TODO: remove hardcoded services
    """

    def __init__(self, storage, services, config=None):
        self.config = config
        self.storage = storage
        self.services = services

    def save_list_status_mapping(self):
        """ TODO: Optimize lookup
        """
        statuses = list(self.services['redmine'].client.statuses())
        lists = list(self.services['trello'].client.lists(self.config.clients.trello.board_id))

        for status in statuses:
            for trello_list in lists:
                if status.name == trello_list.name:
                    self.storage.set_list_or_status_id(redmine_id=status.id, trello_id=trello_list.id)
                    log.debug('Mapped Status %s to Trello', status.name)
                    break
            else:
                log.debug('Status %s not mapped to Trello', status.name)

    def save_user_member_mapping(self):
        users = list(self.services['redmine'].client.users())
        members = list(self.services['trello'].client.members(self.config.clients.trello.board_id))
        for user in users:
            for member in members:
                if user.login == member.username or u'%s %s' % (user.firstname, user.lastname) == member.fullName:
                    self.storage.set_user_or_member_id(redmine_id=user.id, trello_id=member.id)
                    log.debug('Mapped User %s to Trello', user.login)
                    break
            else:
                log.debug('User %s not mapped to Trello', user.login)


# -*- coding: utf-8 -*-
import requests
from logging import getLogger
from dateutil.parser import parse
log = getLogger(__name__)


class ApiException(Exception):
    """ Error, raised by API wrapper """


class MissingProperty(ApiException):
    """ API object is missing property"""

    def __str__(self):
        return 'API object {0} is missing attribute "{1}"'.format(*self.args)

class ApiObject(object):
    date_fields = tuple()

    def __init__(self, client=None, **data):
        """ Abstract API object constructor.

        :param client: API client
        :param data: object internals
        """
        data = self.prepare_dates(data)
        self._data = data
        self.client = client
        self._uid = data.get('name', data.get('subject') or data.get('id'))
        self.validate()
        log.debug('Instantiated API object %s' % repr(self))

    def __getattr__(self, attr):
        try:
            return self._data[attr]
        except KeyError:
            raise MissingProperty(self, attr)

    def get(self, attr, default=None):
        return self._data.get(attr, default)

    def __repr__(self):
        return unicode(self).encode('utf-8')

    def __unicode__(self):
        return u'<{class_name} {identifier}>'.format(class_name=self.__class__.__name__, identifier=self._uid)

    def validate(self):
        """ Validate created object"""

    def prepare_dates(self, data):
        for field in self.date_fields:
            if data.get(field):
                data[field] = parse(data[field])
        return data


class ApiClient(object):
    """ Base class for all service APIs"""

    def default_payload(self):
        return {}

    def build_api_url(self, path):
        raise NotImplemented

    def get(self, path, **payload):
        return self.method('get', path, **payload)

    def post(self, path, data=None, **payload):
        return self.method('post', path, data, **payload)

    def put(self, path, data=None, **payload):
        return self.method('put', path, data, **payload)

    def method(self, method, path, data=None, headers=None, expect_json=True, **payload):
        payload.update(self.default_payload())
        response = requests.request(method=method, url=self.build_api_url(path),
                                    headers=headers, data=data, params=payload)
        log.debug('Request: [%s] %s', response.request.method, response.request.full_url)
        log.debug('Request data: %s', response.request.data)

        try:
            response.raise_for_status()
        except requests.exceptions.HTTPError as e:
            log.error('Request failed: %s', e.response.content)
            raise e

        if response.json is not None:
            return response.json
        else:
            if expect_json:
                raise ApiException('JSON expected, got %s' % response.content)
            return True




# -*- coding: utf-8 -*-

def periodic(scheduler, interval, action, actionargs=()):
    scheduler.enter(interval, 1, periodic,
        (scheduler, interval, action, actionargs))
    action(*actionargs)


# -*- coding: utf-8 -*-
from logging import getLogger
import pprint
from api.trello import Trello
from api.redmine import Redmine
from copydog.api.redmine import Issue
from copydog.api.trello import Card
from copydog.utils import pandoc

log = getLogger('copydog')
pp = pprint.PrettyPrinter(indent=4)


class BaseAdapter(object):
    service_name = None
    writable = False

    def __init__(self, options, storage):
        self.config = options
        self.storage = storage
        self.client = self.get_client(options)

        self.writable = self.config.get('write')
        log.debug('%s is %swritable' % (self.service_name, 'not ' if not self.writable else ''))

    def __str__(self):
        return self.service_name

    def get_client(self, options):
        """ Instantiate remote API
        """
        raise NotImplemented

    def read(self):
        """ Read issues suitable for syncing from client
        """
        last_time_read = self.storage.get_last_time_read(self.service_name)
        log.debug('Reading %s since %s' % (self.service_name, last_time_read))
        issues = self.get_issues_since_last_sync(last_time_read)
        self.mark_read()
        return issues

    def mark_read(self, issue=None):
        self.storage.mark_read(self.service_name, issue=None)

    def get_issues_since_last_sync(self, last_read):
        """ Client-specific API call
        """
        raise NotImplemented

    def convert_to_local_issue(self, foreign_issue):
        raise NotImplemented

    def add_foreign_issue_reference(self, issue, foreign_issue):
        raise NotImplemented

    def write(self, foreign_issue):
        """ Write items from other service to client
        """
        log.debug('Saving issue %s to %s', foreign_issue.id, self.service_name)
        log.debug('%s', pp.pformat(foreign_issue._data))
        local_issue = self.convert_to_local_issue(foreign_issue)
        log.debug('%s', pp.pformat(local_issue))
        local_issue.save()
        if local_issue.is_created():
            self.add_foreign_issue_reference(issue=local_issue, foreign_issue=foreign_issue)
        local_issue.fetch()
        self.storage.mark_written(self.service_name, local_issue, foreign_id=foreign_issue.id)


class RedmineAdapter(BaseAdapter):
    service_name = 'redmine'

    def get_client(self, options):
        return Redmine(host=options['host'], api_key=options['api_key'])

    def get_issues_since_last_sync(self, last_read):
        issues = self.client.issues(
            updated__after=last_read,
            tracker_id=self.config.get('tracker_id'),
            project_id=self.config.get('project_id'),
            fixed_version_id=self.config.get('fixed_version_id')
        )
        return issues

    def convert_to_local_issue(self, foreign_issue):
        assert isinstance(foreign_issue, Card)
        service_from = 'trello'

        if len(foreign_issue.idMembers):
            assigned_to_id = self.storage.get_user_or_member_id(service_from, foreign_issue.idMembers[0])
        else:
            assigned_to_id = None

        issue = Issue(
            id = self.storage.get_opposite_item_id(service_from, foreign_issue.id),
            assigned_to_id = assigned_to_id,
            subject = foreign_issue.name,
            description = pandoc.convert(foreign_issue.desc, 'markdown', 'textile'),
            status_id = self.storage.get_list_or_status_id(service_from, foreign_issue.idList),
            project_id = self.config.project_id,
            tracker_id = self.config.get('tracker_id'),
            due_date = foreign_issue.get('due'),
            client = self.client
        )
        return issue

    def add_foreign_issue_reference(self, issue, foreign_issue):
        """ Currently can't update issue journal to leave a comment
            http://www.redmine.org/issues/10171
        """


class TrelloAdapter(BaseAdapter):
    service_name = 'trello'

    def get_client(self, options):
        return Trello(api_key=options['api_key'], token=options['token'])

    def get_issues_since_last_sync(self, last_read):
        issues = self.client.cards(
            updated__after=last_read,
            board_id=self.config.board_id,
            actions='all')
        return issues

    def convert_to_local_issue(self, foreign_issue):
        assert isinstance(foreign_issue, Issue)
        service_from = 'redmine'
        if hasattr(foreign_issue, 'assigned_to'):
            idMembers = [self.storage.get_user_or_member_id(service_from, foreign_issue.assigned_to['id'])]
        else:
            idMembers = [None]
        card = Card(
            id = self.storage.get_opposite_item_id(service_from, foreign_issue.id),
            idMembers = idMembers,
            name = foreign_issue.subject,
            desc = pandoc.convert(foreign_issue.get('description'), 'textile', 'markdown'),
            idList = self.storage.get_list_or_status_id(service_from, foreign_issue.status['id']),
            idBoard = self.config.board_id,
            due = foreign_issue.get('due_date', 'null'),
            client = self.client,
        )
        return card

    def add_foreign_issue_reference(self, issue, foreign_issue):
        """ Let's create a comment with redmine reference
            TODO: move .post into trello.py
        """
        text = '{service_name}: {issue_url}'.format(
            service_name=foreign_issue.client.service_name,
            issue_url=foreign_issue.get_url()
        )
        self.client.post(path='cards/{card_id}/actions/comments'.format(card_id=issue.id),
            data={'text': text})



# -*- coding: utf-8 -*-
from unittest import TestCase
from mock import patch
from copydog.utils.config import Config
from copydog.watcher import Watch
from copydog.bin.runner import execute


@patch('copydog.bin.runner.Watch', spec=Watch)
@patch('copydog.bin.runner.DaemonRunner')
class TestRunner(TestCase):

    def test_execute_runs_watch_by_default(self, daemon_mock, watch_mock):
        execute(Config(file='tests/test.yml'))
        self.assertTrue(watch_mock.called)
        watch_mock.return_value.run.assert_called_with()

    def test_execute_can_run_daemonized(self, daemon_mock, watch_mock):
        execute(Config(file='tests/test.yml'), daemonize=True)
        self.assertTrue(daemon_mock.called)
        self.assertFalse(watch_mock.return_value.run.called)






# -*- coding: utf-8 -*-
import json
from logging import getLogger
from .base import ApiObject, ApiException, ApiClient
log = getLogger('copydog.api')


class RedmineException(ApiException):
    pass


class Redmine(ApiClient):
    """ Redmine API class """
    service_name = 'redmine'

    def __init__(self, host, api_key=None):
        """ Creates api client instance.

        :host: Full URL of redmine installation.
        :api_key: API key for Redmine. Can be obtained at http://<your_domain>/my/account
        """
        self.host = host.rstrip('/')
        self.api_key = api_key

    def default_payload(self):
        return {'key': self.api_key}

    def build_api_url(self, path):
        return '{host}/{path}.json'.format(host=self.host.strip('/'), path=path)

    def get_many(self, path, **payload):
        while True:
            data = self.method('get', path, **payload)
            for item in data[path]:
                yield item
            total_read = data.get('offset', 0) + data.get('limit', 0)
            if payload.get('limit') or total_read >= data.get('total_count', 0):
                break
            payload['offset'] = total_read

    def post(self, path, data=None, **payload):
        return self.method('post', path, json.dumps(data), headers={'Content-Type': 'application/json'}, **payload)

    def put(self, path, data=None, **payload):
        """ Redmine returns empty 200 OK"""
        return self.method('put', path, json.dumps(data), expect_json=False,
                           headers={'Content-Type': 'application/json'}, **payload)

    def issues(self, inverse=None, updated__after=None, **kwargs):
        """ Issue generator

        :param page: (optional) page number
        :param offset (optional): skip this number of issues in response
        :param limit (optional): number of issues per page
        :param project_id (optional): get issues from the project with the given id,
                           where id is either project id or project identifier
        :param fixed_version_id (optional): target version of the issues as a filter
        :param tracker_id (optional): get issues from the tracker with the given id
        :param updated__after (optional): only issues, update after given timestamp
        :param sort (optional): column to sort with
        :param inverse (optional): inverse sorting
        :type inverse: boolean

        Ref: http://www.redmine.org/projects/redmine/wiki/Rest_Issues
        """
        if inverse and kwargs.get('sort'):
            kwargs['sort'] += ':desc'
        if updated__after:
            kwargs['updated_on'] = '>={0}'.format(updated__after.date().isoformat())

        for data in self.get_many('issues', **kwargs):
            issue = Issue(self, **data)
            # Redmine doesn't allow granular search by timestamp, so filtering manually
            log.debug('Issue comparing %s > %s' % (issue.updated_on, updated__after))
            if updated__after and issue.updated_on <= updated__after:
                log.debug('ignoring')
                continue
            log.debug('passed')
            yield issue

    def projects(self):
        """ Get a list of projects
        """
        for data in self.get_many('projects'):
            yield Project(self, **data)

    def trackers(self):
        """ Get a list of trackers
        """
        for data in self.get_many('trackers'):
            yield Tracker(self, **data)

    def statuses(self):
        """ Get a list of statuses

        Ref: http://www.redmine.org/projects/redmine/wiki/Rest_IssueStatuses
        """
        for data in self.get_many('issue_statuses'):
            yield Status(self, **data)

    def users(self):
        """ Get a list of all users

        Ref: http://www.redmine.org/projects/redmine/wiki/Rest_Users
        """
        for data in self.get_many('users'):
            yield User(self, **data)


class Project(ApiObject):
    """ Redmine project"""


class Tracker(ApiObject):
    """ Redmine tracker"""


class Status(ApiObject):
    """ Redmine status"""


class User(ApiObject):
    """ Redmine user"""


class Issue(ApiObject):
    """ Redmine issue

        :param id: issue's unique id
        :param subject: issue name
        :param description: description
    """
    created = False
    date_fields = ('updated_on', 'created_on')

    def get_url(self):
        return '{host}/issues/{issue_id}/'.format(host=self.client.host, issue_id=self.id)

    def save(self):
        """ Save new issue

        Redmine expects:
        {
            "issue": {
                "project_id": "example",
                "subject": "Test issue",
                "custom_field_values":{
                    "1":"1.1.3"  #the affected version field
                }
            }
        }
        """
        if self.get('id'):
            result = self.client.put(path='issues/{issue_id}'.format(
                issue_id=self.id),
                data={'issue': self._data})
        else:
            result = self.client.post(path='issues', data={'issue': self._data})
            self._data = result['issue']
            self.created = True
        return self

    def fetch(self):
        """ Fetch fresh info about the issue

        We need it, because save method doesn't return card timestamp on PUT.
        """
        result = self.client.get('issues/{issue_id}'.format(issue_id=self.id))
        self._data = result['issue']
        return self

    @property
    def last_updated(self):
        return self.updated_on

    def is_created(self):
        return self.created



from setuptools import setup
from copydog import __version__


setup(
    name='copydog',
    version=__version__,
    packages=['copydog', 'copydog.api', 'copydog.bin',
              'copydog.storage', 'copydog.storage.backends', 'copydog.utils'],
    install_requires=[
        'requests==0.13.6',
        'redis==2.6.2',
        'python-dateutil==2.1',
        'pytz==2012c',
        'docopt==0.4.1',
        'PyYAML==3.10',
        'python-daemon==1.6',
        'texttable==0.8.1',
    ],
    entry_points={
        "console_scripts": [
            'copydog = copydog.bin.runner:main'
    ]},
    url='http://copydog.readthedocs.org/',
    license='BSD',
    author='coagulant',
    author_email='baryshev@gmail.com',
    description='Copies issues between Redmine and Trello on the fly',
    long_description=open('README.rst').read(),
    classifiers=(
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
    )
)



# -*- coding: utf-8 -*-
from texttable import Texttable


def print_table(prefix, items):
    table = Texttable(max_width=160)
    table.set_deco(Texttable.HEADER)
    table.header(['%s_id' % prefix, '%s_updated' % prefix, '%s_fk' % prefix])
    for key, values in items.iteritems():
        table.add_row([key, values.get('updated'), values.get('opposite_id')])
    print table.draw() + "\n"


def print_mapping(prefix, key, items):
    table = Texttable(max_width=160)
    table.set_deco(Texttable.HEADER)
    table.header(['%s_%s' % (prefix, key),  '%s_fk' % prefix])
    for key, value in items.iteritems():
        table.add_row([key, value])
    print table.draw() + "\n"


def browse(storage):
    redmine_keys = storage.redis.keys(pattern='redmine:items*')
    trello_keys = storage.redis.keys(pattern='trello:items*')
    redmine_list_mapping = storage.redis.hgetall('redmine:list_status_mapping')
    trello_list_mapping = storage.redis.hgetall('trello:user_member_mapping')
    redmine_user_mapping = storage.redis.hgetall('redmine:user_member_mapping')
    trello_user_mapping = storage.redis.hgetall('trello:user_member_mapping')

    issues = dict((key, storage.redis.hgetall(key)) for key in redmine_keys)
    cards = dict((key, storage.redis.hgetall(key)) for key in trello_keys)

    print 'Copydog storage'
    print '==============='

    print 'Redmine last read at %s' % storage.get_last_time_read('redmine')
    print 'Trello last read at %s \n' % storage.get_last_time_read('trello')

    print_mapping('redmine', 'list_status_id', redmine_list_mapping)
    print_mapping('trello', 'list_status_id', trello_list_mapping)

    print_mapping('redmine', 'user_member_id', redmine_user_mapping)
    print_mapping('trello', 'user_member_id', trello_user_mapping)

    print_table('redmine', issues)
    print_table('trello', cards)



# -*- coding: utf-8 -*-
from logging import getLogger
from dateutil.parser import parse
from .base import ApiObject, ApiException, ApiClient
log = getLogger('copydog.api')


class TrelloException(ApiException):
    pass


class Trello(ApiClient):
    """ Trello API class """
    service_name = 'trello'

    def __init__(self, api_key=None, token=None):
        """ Creates api client instance.

        :param api_key: Developer API key for Trello. Can be obtained at https://trello.com/1/appKey/generate
        :param token: Authorization token with read or read-write access for some period of time
                https://trello.com/docs/gettingstarted/index.html#token
        """
        self.api_key = api_key
        self.token = token
        self.host = 'https://api.trello.com/1/'

    def default_payload(self):
        return dict(key=self.api_key, token=self.token)

    def build_api_url(self, path):
        return '{host}/{path}'.format(host=self.host.strip('/'), path=path)

    def get_many(self, path, **payload):
        json = self.method('get', path, **payload)
        for item in json:
            yield item

    def boards(self):
        """ Get a list of boards
        """
        for data in self.get_many('members/me/boards/'):
            yield Board(self, **data)

    def lists(self, board_id):
        """ Get list of lists :)
        """
        for data in self.get_many('boards/{board_id}/lists'.format(board_id=board_id)):
            yield List(self, **data)

    def members(self, board_id):
        """ Get list of lists :)
        """
        for data in self.get_many('boards/{board_id}/members'.format(board_id=board_id)):
            yield Member(self, **data)

    def cards(self, board_id, **kwargs):
        """ Get a list of cards

        :param board_id: The id of board to look for cards
        """
        updated__after = kwargs.pop('updated__after', None)
        num_cards_recieved = 0

        for data in self.get_many('boards/{board_id}/cards'.format(board_id=board_id), **kwargs):
            num_cards_recieved += 1
            card = Card(self, **data)
            if updated__after and card.last_updated <= updated__after:
                continue
            yield card

        log.debug('Got whole lot of %s cards from Trello board', num_cards_recieved)


class Board(ApiObject):
    """ Trello board """
    pass


class List(ApiObject):
    """ Trello list """
    pass


class Member(ApiObject):
    """ Trello board """
    pass


class Card(ApiObject):
    """ Trello card

        :param id: (optional) card's unique hash
        :param name: card's name
        :param desc: (optional) description
        :param idList: list id
        :param url: (optional) URL
    """
    created = False

    def validate(self):
        assert self.name is not None
        assert self.idList is not None

    def get_url(self):
        return self.url

    def save(self):
        """ Save new card
        """
        if self.get('id'):
            result = self.client.put(path='cards/{card_id}'.format(card_id=self.id), data=self._data)
        else:
            result = self.client.post(path='cards', data=self._data)
            # Trello doesn't allow to assign cards on creation
            if self.idMembers[0]:
                self.client.post(path='cards/{card_id}/members'.format(card_id=result['id']),
                                 data={'value': self.idMembers[0]})
                result['idMembers'] = self.idMembers
            self.created = True
        self._data = result
        return self

    def fetch(self):
        """ Fetch fresh info about the card

        We need it, because save method doesn't return card timestamp.
        """
        result = self.client.get('cards/{card_id}'.format(card_id=self.id), actions='all')
        self._data = result
        return self

    @property
    def last_updated(self):
        return parse(self.actions[0]['date'])

    def is_created(self):
        return self.created




# -*- coding: utf-8 -*-
import datetime
import os
import types
from unittest import TestCase
from dateutil.tz import tzoffset
from nose.plugins.attrib import attr
from copydog.api.redmine import Redmine, Issue, Project, Status


@attr('http', 'redmine')
class TestRedmine(TestCase):
    """ Make sure to provide ``REDMINE_HOST`` and ``REDMINE_API_KEY``
        environment variables to execute this test.
    """

    def setUp(self):
        self.r = Redmine(host=os.environ.get('COPYDOG_REDMINE_HOST'),
                         api_key=os.environ.get('COPYDOG_REDMINE_API_KEY'))

    def test_connect_public_api(self):
        r = Redmine(host='http://redmine.org')
        issues = r.issues()
        self.assertTrue(isinstance(issues, types.GeneratorType))
        self.assertTrue(next(issues), Issue)

    def test_connect_protected_api(self):
        issues = self.r.issues()
        self.assertTrue(next(issues), Issue)

    def test_projects(self):
        projects = self.r.projects()
        self.assertTrue(type(next(projects)) is Project)

    def test_issues_limit(self):
        issues = self.r.issues(limit=1)
        self.assertTrue(len(list(issues)) <= 1)

    def test_statuses(self):
        statuses = self.r.statuses()
        self.assertTrue(type(next(statuses, Status())) is Status)


@attr('redmine')
class TestIssue(TestCase):

    def test_init(self):
        json = {u'status': {u'name': u'Assigned', u'id': 2},
                u'due_date': u'2012/07/23',
                u'description': u"Issue's description",
                u'author': {u'name': u'\u0418\u043b\u044c\u044f \u0411\u0430\u0440\u044b\u0448\u0435\u0432',
                            u'id': 3},
                u'project': {u'name': u'Test', u'id': 40},
                u'created_on': u'2012/07/22 16:17:37 +0400',
                u'tracker': {u'name': u'\u0420\u0430\u0437\u0440\u0430\u0431\u043e\u0442\u043a\u0430',
                             u'id': 8},
                u'custom_fields': [{u'name': u'\u041f\u0440\u0438\u0451\u043c\u043e\u0447\u043d\u044b\u0439 \u0442\u0435\u0441\u0442',
                                    u'value': u'', u'id': 3}],
                u'assigned_to': {u'name': u'\u0418\u043b\u044c\u044f \u0411\u0430\u0440\u044b\u0448\u0435\u0432',
                                 u'id': 3},
                u'updated_on': u'2012/07/22 21:56:52 +0400',
                u'subject': u'First issue',
                u'id': 6844,
                u'done_ratio': 0,
                u'priority': {u'name': u'\u041d\u043e\u0440\u043c\u0430\u043b\u044c\u043d\u044b\u0439', u'id': 4}}
        self.assertEqual(Issue(**json).created_on, datetime.datetime(2012, 7, 22, 16,17,37, tzinfo=tzoffset(None, 14400)))


# -*- coding: utf-8 -*-
from abc import ABCMeta, abstractmethod


class BaseStorage(object):
    """ Abstract storage class"""

    __metaclass__ = ABCMeta

    @abstractmethod
    def flush(self):
        """ Empty storage, as if copydog never launched.
            Handly for debugging and cleaning up the mess.
        """

